/* Secure RM - by van Hauser / [THC], vh@reptile.rug.ac.be
 *
 * Secure ReMove first overwrites then renames and finally deletes the target
 * file(s) specified via parameters.
 * For security reasons full 32kb blocks are written so that the whole block
 * on which the file(s) live are overwritten. (change #define #BLOCKSIZE)
 * The option -l overwrites two times the data.
 * The option -ll overwrites the data once.
 * Standard mode is a real security wipe for 38 times, flushing
 * the caches after every write. The wipe technique was proposed by Peter
 * Gutmann at Usenix '96 and includes 10 random overwrites plus 28 special
 * defined characters. Take a look at the paper of him, it's really worth
 * your time. 
 *
 * Advice : set "alias rm 'srm -v'"
 *
 * Read the manual for limitations.
 * Compiles clean on OpenBSD, Linux, Solaris and AIX
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <dirent.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/resource.h>
   

#define RANDOM_DEVICE	"/dev/urandom"	/* must not exist */
#define BLOCKSIZE       32769		/* must be mod 3 = 0, should be >= 16k */
#define DIR_SEPERATOR	'/'
#define FLUSH		sync()

#ifndef O_SYNC
 #ifdef O_FSYNC
  #define O_SYNC O_FSYNC
 #else
  #define O_SYNC 0
 #endif
#endif

char *VERSION = "v2.3 by van Hauser/THC <vh@reptile.rug.ac.be> http://www.infowar.co.uk/thc";
unsigned long bufsize = BLOCKSIZE;
char buf[65535];
char newname[512];
char rmfile[200];
char *prog;
FILE *devrandom = NULL;
FILE *f;
int file;
int slow = O_SYNC;
int verbose = 0;
int recursive = 0;

void help () {
    printf("%s %s\n\n",prog, VERSION);
    printf("Syntax: %s [-dflv] file1 file2 etc.\n\n", prog);
    printf("%s does a secure overwrite/rename/delete of the target file(s).\n", prog);
    printf("Standard is secure mode (38 writes).\n");
    printf("Options:\n");
    printf("\t-d  ignore the two dot special files \".\" and \"..\".\n");
    printf("\t-f  fast (and insecure mode): no /dev/urandom, no synchronize mode.\n");
    printf("\t-l  lessens the security (use twice for total insecure mode).\n");
    printf("\t-v  is verbose mode.\n");
    exit(1);
}

void fill_buf(char pattern[3]) {
    int loop;
    int where;
    
    for (loop = 0; loop < (bufsize / 3); loop++) {
        where = loop * 3;
	buf[where] = pattern[0];
	buf[where+1] = pattern[1];
	buf[where+2] = pattern[2];
    }
}

void random_buf() {
    int loop;
    
    if (devrandom == NULL)
        for (loop = 0; loop < bufsize; loop++)
            buf[loop] = (unsigned char) (256.0*rand()/(RAND_MAX+1.0));
    else
        fread(&buf, bufsize, 1, devrandom);
}

int smash_it(char *filename, int mode)  {
    unsigned char write_modes[27][3] = {
        {"\x55\x55\x55"}, {"\xaa\xaa\xaa"}, {"\x92\x49\x24"}, {"\x49\x24\x92"},
        {"\x24\x92\x49"}, {"\x00\x00\x00"}, {"\x11\x11\x11"}, {"\x22\x22\x22"},
        {"\x33\x33\x33"}, {"\x44\x44\x44"}, {"\x55\x55\x55"}, {"\x66\x66\x66"},
        {"\x77\x77\x77"}, {"\x88\x88\x88"}, {"\x99\x99\x99"}, {"\xaa\xaa\xaa"},
        {"\xbb\xbb\xbb"}, {"\xcc\xcc\xcc"}, {"\xdd\xdd\xdd"}, {"\xee\xee\xee"},
        {"\xff\xff\xff"}, {"\x92\x49\x24"}, {"\x49\x24\x92"}, {"\x24\x92\x49"},
        {"\x6d\xb6\xdb"}, {"\xb6\xdb\x6d"}, {"\xdb\x6d\xb6"}
    };

    unsigned char std_array[3] = "\xff\xff\xff";
    unsigned long writes;
    unsigned long counter;
    unsigned long filesize;
    struct stat filestat;
    struct stat controlstat;
    int turn;
    int result;
    int i;
    int i_am_a_directory = 0;

/* get the file stats */
    if (lstat(filename, &filestat))
            return 1;

/* if the blocksize on the filesystem is bigger than the on compiled with, enlarge! */
    if (filestat.st_blksize > bufsize) {
	if (filestat.st_blksize > 65532) {
	    bufsize = 65535;
	} else {
	    bufsize = (((filestat.st_blksize / 3) + 1) * 3);
	}
    }

/* handle the recursive mode */
    if (recursive)
        if (S_ISDIR(filestat.st_mode)) {
	    DIR *dir;
	    struct dirent *dir_entry;
	    struct stat cwd_stat;
	    char current_dir[4097];
	    int res;
	    int chdir_success = 1;
 
            if (verbose) printf("DIRECTORY (going recursive now)\n");
            getcwd(current_dir, 4096);
            current_dir[4096] = '\0';
            
            /* a won race will chmod a file to 0700 if the user is owner/root
               I'll think about a secure solution to this, however, I think
               there isn't one - anyone with an idea? */
            if (chdir(filename)) {
                (void) chmod(filename, 0700); /* ignore permission errors */
                if (chdir(filename)) {
                    fprintf(stderr,"Can't chdir() to %s, hence I can't wipe it.\n", filename);
                    chdir_success = 0;
                }
            }

	    if (chdir_success) {
	        lstat(".", &controlstat);
	        lstat("..", &cwd_stat);
	        if ( (filestat.st_dev != controlstat.st_dev) || (filestat.st_ino != controlstat.st_ino) ) {
	            fprintf(stderr, "Race found! (directory %s became a link)\n", filename);
	        } else {
  	            if ((dir = opendir (".")) != NULL) {
	                (void) chmod(".", 0700); /* ignore permission errors */
	                dir = opendir (".");
	            }
		    if (dir != NULL) {
		        while ((dir_entry = readdir(dir)) != NULL)
			    if (strcmp(dir_entry->d_name, ".") && strcmp(dir_entry->d_name, "..")) {
			        if (verbose) printf("Wiping %s ", dir_entry->d_name);
		  	        if ( (res = smash_it(dir_entry->d_name, mode)) > 0) {
		  	            if (res == 3)
		  	                fprintf(stderr,"File %s was raced, hence I won't wipe it.\n", dir_entry->d_name);
		  	            else {
			                fprintf(stderr,"Couldn't delete %s. ", dir_entry->d_name);
			                perror("");
			            }
			        } else
			            if (verbose) printf(" Done\n");
			    }
		        closedir(dir);
		    }
		}
		if(chdir(current_dir) != 0) {
		    perror("chdir error (aborting)");
		    exit(1);
		}
		lstat(current_dir, &controlstat);
		if ( (cwd_stat.st_dev != controlstat.st_dev) || (cwd_stat.st_ino != controlstat.st_ino) ) {
                    fprintf(stderr, "Race found! (directory %s was exchanged)\n", current_dir);
                    exit(1);
                }
		i_am_a_directory = 1;
	    }
        }
/* end of recursive function */


 if (S_ISREG(filestat.st_mode)) {

/* open the file for writing in sync. mode */
    if ((file = open(filename, O_RDWR | slow)) < 0) {
        /* here again this has a race problem ... hmmm */
        /* make it writable for us if possible */
        (void) chmod(filename, 0600);   /* ignore errors */
        if ((file = open(filename, O_RDWR | slow)) < 0)
    	    return 1;
    }
    if ((f = fdopen(file, "r+b")) == NULL)
        return 1;

    fstat(file, &controlstat);
    if ( (filestat.st_dev != controlstat.st_dev) || (filestat.st_ino != controlstat.st_ino) || (! S_ISREG(controlstat.st_mode)) ) {
        fclose(f);
        return 3;
    }

/* calculate the number of writes */
    filesize = filestat.st_size;
    writes = (1 + (filesize / bufsize));

    if (mode != 0) {
        fill_buf(std_array);
        for (counter=1; counter<=writes; counter++)
            fwrite(&buf, 1, bufsize, f);
        if (verbose)
            printf("*");
        fflush(f);
        if (fsync(file) < 0)
            FLUSH;
    }

/* do the overwriting stuff */
    for (turn=0; turn<=36; turn++) {
        rewind(f); 
        if ((mode < 2) && (turn > 0)) break;
        if ((turn>=5) && (turn<=31)) {
            fill_buf(write_modes[turn-5]);
            for (counter=1; counter<=writes; counter++)
                fwrite(&buf, 1, bufsize, f);
        } else {
	    for (counter=1; counter<=writes; counter++) {
	        random_buf();
	        fwrite(&buf, 1, bufsize, f);
	    }
        }  
        fflush(f);
        if (fsync(file) < 0)
            FLUSH;
        if (verbose)
            printf("*");
    }

    (void) fclose(f);
/* Hard Flush -> Force cached data to be written to disk */
    FLUSH;
/* open + truncating the file, so an attacker doesn't know the diskblocks */
   if ((file = open(filename, O_WRONLY | O_TRUNC | slow)) >= 0)
	close(file);

 } /* end IS_REG() */
 else {
   if (S_ISDIR(filestat.st_mode) && i_am_a_directory == 0) {
        fprintf(stderr,"Warning: %s is a directory. I will not remove it, because the -r option is missing!\n", filename);
        return;
   }
   if (! i_am_a_directory) {
	fprintf(stderr,"Warning: %s is not a regular file, rename/unlink only!", filename);
	if (! verbose)
	    printf("\n");
   }
 }

/* Generate random unique name, renaming and deleting of the file */
    strncpy(newname, filename, sizeof(newname)-1);
    newname[sizeof(newname)-1] = '\0';
    turn = 0;

    do {
        for (i=strlen(newname)-1; (newname[i] != DIR_SEPERATOR)&&(i>=0); i--)
             if (newname[i] != '.') /* keep dots in the filename */
                newname[i] = 97+(int) (27.0*rand()/(RAND_MAX+1.0));
        result = lstat(newname, &filestat);
        turn++;
    } while ((result >= 0)&&(turn<=100));
    if (turn<=100) {
       result = rename(filename, newname);
       if (result !=0) {
          fprintf(stderr,"Warning: Couldn't rename %s. ", filename);
          perror("");
          strncpy(newname, filename, sizeof(newname)-1);
          newname[sizeof(newname)-1] = '\0';
       }
    } else {
       fprintf(stderr,"Warning: Couldn't find a free filename for %s!\n",filename);
       strncpy(newname, filename, sizeof(newname)-1);
       newname[sizeof(newname)-1] = '\0';
    }
    if (i_am_a_directory) {
        result = rmdir(newname);
	if (verbose)
	    printf("Removing DIRECTORY %s:", filename);
	if (result)
	    (void) rename(newname, filename);
    } else
        result = unlink(newname);

    newname[0] = '\0';
    if (result != 0)
        return 2;
    return 0;
}

void cleanup(int signo) {
    fprintf(stderr,"Terminated by signal. Clean exit.\n");
    if (f) {
        fflush(f);
        fsync(file);
        (void) fclose(f);
    }
    if (file >= 0)
    	close(file);
    if (newname[0] != '\0')
        if (unlink(newname) != 0 )
            fprintf (stderr,"Warning: Can't remove %s!\n",newname);
    if (rmfile[0] != '\0')
	fprintf (stderr,"Warning: I won't remove %s, because the secure deletion wasn't completed!\n",rmfile);
    FLUSH;
    fflush(stdout);
    fflush(stderr);
    exit(1);
}

int main (int argc, char *argv[]) {

    int errors = 0;
    int dot = 0;
    int result;
    int secure = 2; /* Standard is now SECURE mode (38 overwrites) [since v2.0] */
    int loop;
    struct rlimit rlim;

    prog = argv[0];
    if (argc < 2)
        help();

    while (1) {
        result = getopt(argc, argv, "DdFfLlRrSsVv");
        if (result<0) break;
        switch (result) {
            case 'd' :
            case 'D' : dot = 1;
                       break;
            case 'F' :
            case 'f' : slow = 0;
            	       break;
	    case 'L' :
	    case 'l' : if (secure) secure--;
	    	       break;
	    case 'R' :
	    case 'r' : recursive++;
	    	       break;
	    case 'S' :
            case 's' : secure++;
            	       break;
	    case 'V' :
            case 'v' : verbose++;
            	       break;
            default :  help();       
        }
    }
    loop = optind;
    if (loop == argc)
        help();
    if (BLOCKSIZE<16384)
        fprintf(stderr, "Warning: incompiled blocksize is <16k !\n");
    if (BLOCKSIZE % 3 > 0)
        fprintf(stderr, "Warning: incompiled blocksize is not a multiple of 3!\n");
    (void) setvbuf(stdout, NULL, _IONBF, 0);
    srand( (getpid()+getuid()+getgid()) ^ time(0) );

    signal(SIGINT, cleanup);
    signal(SIGTERM, cleanup);
    signal(SIGHUP, cleanup);

    if (verbose)
        switch (secure) {
           case 0 : printf("Wipe mode is insecure (one pass [random])\n");
		    break;
	   case 1 : printf("Wipe mode is insecure (two passes [0xff/random])\n");
	   	    break;
	   default: printf("Wipe mode is secure (38 special passes)\n");
	}
    if (slow && secure)
        if ((devrandom = fopen(RANDOM_DEVICE, "r")) != NULL)
            if (verbose)
                printf("Using %s for random input.\n", RANDOM_DEVICE);

#ifdef RLIM_INFINITY
    rlim.rlim_cur = RLIM_INFINITY;
    rlim.rlim_max = RLIM_INFINITY;
 #ifdef RLIMIT_FSIZE
    if (setrlimit(RLIMIT_FSIZE, &rlim) != 0)
    	fprintf(stderr, "Warning: Could not reset ulimit for filesize.\n");
 #endif
#endif

    while (loop < argc) {
        strncpy(rmfile, argv[loop], sizeof(rmfile)-1);
        rmfile[sizeof(rmfile)-1] = '\0';
        loop++;
        if (strcmp("/", rmfile) == 0) {
            fprintf(stderr,"Warning: Do you really want to remove the ROOT directory??\n");
            fprintf(stderr,"I'm giving you 5 seconds to abort ... press Control-C\n");
	    sleep(6);
            fprintf(stderr,"Doing my evil work now, don't whimp later, you had been informed.\n");
        }
        if (dot)
            if ((strcmp(".", rmfile) == 0) || (strcmp("..", rmfile) == 0))
                continue;
        if (verbose) printf("Wiping %s ", rmfile);
        result = (int) smash_it(rmfile, secure);
        switch (result) {
            case 0 : if (verbose) printf(" Done\n");
                     break;
            case 1 : perror(rmfile);
            	     break;
            case 2 : fprintf(stderr,"Couldn't delete %s. ", rmfile);
            	     perror("");
                     break;
            case 3 : fprintf(stderr,"File %s was raced, hence I won't wipe it.\n", rmfile);
                     break;
            default: fprintf(stderr,"Unknown error\n");
        }
        if (result)
            errors++;
    }
    if (devrandom)
        fclose(devrandom);
    if (errors)
        return 1;
    else
        return 0;
}
