/* Secure SWAP cleaner - by van Hauser / [THC], vh@reptile.rug.ac.be
 *
 * Note that this program is beta. It was tested with linux, but I can't
 * tell for other platforms. Read the statement at #define SWAP_PAGESIZE
 * on how to use this program on other unix machines.
 *
 * of course: you have to turn of the swapspace before using this program !
 *
 * Secure SWAP overwrites all data on your swap device.
 * Standard mode is a real security wipe for 38 times, flushing
 * the caches after every write. The wipe technique was proposed by Peter
 * Gutmann at Usenix '96 and includes 10 random overwrites plus 28 special
 * defined characters. Take a look at the paper of him, it's really worth
 * your time.
 * The option -l overwrites two times the data. (0xff + random)
 * The option -ll overwrites the data once. (random)
 *
 * Read the manual for limitations.
 * Compiles clean on OpenBSD, Linux, Solaris and AIX
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <strings.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

/* SWAP_PAGESIZE is an important variable. You have to set this
 * to your header length of your swapdevice. For Linux this is 4096,
 * I don't know for the other OSs. To be sure, set this to 0 and
 * recreate your swapspace afterwards (for linux: mkswap /dev/swapdevice)
 */
#define SWAP_PAGESIZE	4096
#define FLUSH		sync()
#define RANDOM_DEVICE	"/dev/urandom"  /* must not exist */
#define BLOCKSIZE	131073

#ifndef O_SYNC
 #ifdef O_FSYNC
  #define O_SYNC O_FSYNC
 #else
  #define O_SYNC 0
 #endif
#endif

char *VERSION = "v2.3 by van Hauser/THC <vh@reptile.rug.ac.be> http://www.infowar.co.uk/thc";
char *prog;
char buf[BLOCKSIZE];
int verbose;
int fd;
int slow = O_SYNC;
FILE *file;
FILE *devrandom = NULL;

void help() {
    printf("%s %s\n\n",prog, VERSION);
    printf("Syntax: %s [-flv] /dev/of_swap_device\n\n", prog);
    printf("%s does a secure overwrite of the swap space\n", prog);
    printf("Standard is secure mode (38 writes).\n");
    printf("Options:\n");
    printf("\t-f  fast (and insecure mode): no /dev/urandom, no synchronize mode.\n");
    printf("\t-l  lessens the security (use twice for total insecure mode).\n");
    printf("\t-v  is verbose mode.\n");
    printf("\nYou must disable the swapspace before using this program! This is *BETA*!\n");
    exit(1);
}

void fill_buf(char pattern[3]) {
    int loop;
    int where;

    for (loop = 0; loop < (BLOCKSIZE / 3); loop++) {
        where = loop * 3;
       buf[where] = pattern[0];
       buf[where+1] = pattern[1];
       buf[where+2] = pattern[2];
    }
}

void random_buf() {
    int loop;

    if (devrandom == NULL)
        for (loop = 0; loop < BLOCKSIZE; loop++)
            buf[loop] = (unsigned char) (256.0*rand()/(RAND_MAX+1.0));
    else
        fread(&buf, BLOCKSIZE, 1, devrandom);
}

int smash_it(char *filename, int mode) {
    unsigned char write_modes[27][3] = {
        {"\x55\x55\x55"}, {"\xaa\xaa\xaa"}, {"\x92\x49\x24"}, {"\x49\x24\x92"},
        {"\x24\x92\x49"}, {"\x00\x00\x00"}, {"\x11\x11\x11"}, {"\x22\x22\x22"},
        {"\x33\x33\x33"}, {"\x44\x44\x44"}, {"\x55\x55\x55"}, {"\x66\x66\x66"},
        {"\x77\x77\x77"}, {"\x88\x88\x88"}, {"\x99\x99\x99"}, {"\xaa\xaa\xaa"},
        {"\xbb\xbb\xbb"}, {"\xcc\xcc\xcc"}, {"\xdd\xdd\xdd"}, {"\xee\xee\xee"},
        {"\xff\xff\xff"}, {"\x92\x49\x24"}, {"\x49\x24\x92"}, {"\x24\x92\x49"},
        {"\x6d\xb6\xdb"}, {"\xb6\xdb\x6d"}, {"\xdb\x6d\xb6"}
    };

    unsigned char std_array[3] = "\xff\xff\xff";
    int turn;
    struct stat stats;

    if ((fd = open (filename, O_RDWR | slow)) < 0) {
        fprintf(stderr, "can't open %s for writing.", filename);
        perror("");
        exit(1);
    }
    file = fdopen(fd, "r+b");
    stat(filename, &stats);
    if (!S_ISBLK(stats.st_mode)) {
        fprintf(stderr, "not a block device : %s\n", filename);
        exit(1);
    }

    if (BLOCKSIZE % 3 > 0)
        fprintf(stderr, "Warning: incompiled blocksize is not a multiple of 3!\n");

    if (verbose) {
        switch (mode) {
           case 0 : printf("Wipe mode is insecure (one pass [random])\n");
                    break;
           case 1 : printf("Wipe mode is insecure (two passes [0xff/random])\n");                    break;
           default: printf("Wipe mode is secure (38 special passes)\n");
        }
 	printf("Writing to swap device %s: ", filename);
    }

    if (slow && mode)
        if ((devrandom = fopen(RANDOM_DEVICE, "r")) != NULL)
            if (verbose)
                printf("Using %s for random input.\n", RANDOM_DEVICE);

    if (fseek(file, SWAP_PAGESIZE, SEEK_SET) < 0) {
        fprintf(stderr,"can't seek to offset %d on %s\n",SWAP_PAGESIZE,filename);
        exit(1);
    }


    if (mode > 0) {
        fill_buf(std_array);
        do {} while (fwrite(&buf, 1, BLOCKSIZE, file) == BLOCKSIZE);
        if (verbose)
            printf("*");
        fflush(file);
    }

        for (turn=0; turn<=36; turn++) {
            if (fseek(file, SWAP_PAGESIZE, SEEK_SET) < 0) {
                fprintf(stderr,"can't seek to offset %d on %s\n",SWAP_PAGESIZE,filename);
                exit(1);
            }
            if ((mode < 2) && (turn > 0)) break;
            if ((turn>=5) && (turn<=31)) {
                fill_buf(write_modes[turn-5]);
                do {} while (fwrite(&buf, 1, BLOCKSIZE, file) == BLOCKSIZE);
            } else {
                do {
                    random_buf();
                } while (fwrite(&buf, 1, BLOCKSIZE, file) == BLOCKSIZE);
    	    }
    	    FLUSH;
    	    if (verbose)
    	        printf("*");
    	}
    
    fclose(file);
    if (devrandom)
        fclose(devrandom);
    FLUSH;
    
    if (verbose)
	printf(" done\n");
    
    return 0;
}

void cleanup() {
    fprintf(stderr,"Terminated by signal. Clean exit.\n");
    if (file) {
        fflush(file);
        FLUSH;
        fclose(file);
    }
    FLUSH;
    fflush(stdout);
    fflush(stderr);
    exit(1);
}

int main (int argc, char *argv[]) {
    int secure = 2;
    int result;
    
    prog = argv[0];
    if (argc==1)
        help();

    while (1) {
        result = getopt(argc, argv, "FfLlSsVv");
        if (result<0) break;
        switch (result) {
            case 'F' :
            case 'f' : slow = 0;
                       break;
            case 'L' :
            case 'l' : if (secure) secure--;
            	       break;
            case 'S' :
            case 's' : secure++;
                       break;
            case 'V' :
            case 'v' : verbose++;
                       break;
            default :  help();
        }
    }

    if ((optind+1) != argc)
        help();
    (void) setvbuf(stdout, NULL, _IONBF, 0);
    signal(SIGINT, cleanup);
    signal(SIGTERM, cleanup);
    signal(SIGHUP, cleanup);

    smash_it(argv[optind], secure);
    /* thats all */
    exit(0);
}
